/* maze generator by Sven Koenig and William Yeoh (USC), inspired by Joseph Pemberton */

#include <stdio.h>
#include <stdlib.h>

#define MAZEWIDTH  101            /* the width of the maze (must be an odd number)                                       */
#define MAZEHEIGHT 101            /* the height of the maze (must be an odd number)                                      */
#define MAZEDENSITY 0.25          /* percentage of blocked cells if RANDOMMAZE is defined                                */
#define WALLSTOREMOVE 100         /* number of walls to remove if RANDOMMAZE is NOT defined - infinite loop if too large */
#define RANDOMSTARTGOAL           /* whether the start and goal state are drawn randomly                                 */
#define STARTX 1                  /* x coordinate of the start cell                                                      */
#define STARTY 1                  /* y coordinate of the start cell                                                      */
#define GOALX 100                 /* x coordinate of the goal  cell                                                      */
#define GOALY 100                 /* y coordinate of the goal  cell                                                      */

#define DIRECTIONS 4
static int dx[DIRECTIONS]      = {1,  0, -1,  0};
static int dy[DIRECTIONS]      = {0,  1,  0, -1};

int goaly = GOALY;
int goalx = GOALX;
int starty = STARTY;
int startx = STARTX;

struct cell;
typedef struct cell cell;

struct cell
{
  short obstacle;
  int dfsx, dfsy; 
  cell *succ[DIRECTIONS];
};

cell **maze;
cell *mazestart, *mazegoal;

void preprocessmaze()
{
  int x, y, d;
  int newx, newy;
  
  if (maze == NULL){
    maze = (cell **)calloc(MAZEHEIGHT, sizeof(cell *));
    for (y = 0; y < MAZEHEIGHT; ++y){
      maze[y] = (cell *)calloc(MAZEWIDTH, sizeof(cell));
      for (x = 0; x < MAZEWIDTH; ++x){
	for (d = 0; d < DIRECTIONS; ++d){
	  newy = y + dy[d];
	  newx = x + dx[d];
	  maze[y][x].succ[d] = (newy >= 0 && newy < MAZEHEIGHT && newx >= 0 && newx < MAZEWIDTH) ? &maze[newy][newx] : NULL;
	}
      }
    }
  } 
#ifdef RANDOMSTARTGOAL
  goaly = (random() % ((MAZEHEIGHT + 1) / 2 )) * 2;
  goalx = (random() % ((MAZEWIDTH  + 1) / 2 )) * 2;
  while (1){
    starty = (random() % ((MAZEHEIGHT + 1) / 2 )) * 2;
    startx = (random() % ((MAZEWIDTH  + 1) / 2 )) * 2;
    if (startx != goalx || starty != goaly)
      break;
  }
#endif
  mazegoal = &maze[goaly][goalx];
  mazestart =  &maze[starty][startx];   
}

void newdfsmaze(int wallstoremove)
{
    int d, dtmp;
    int x, y;
    int newx, newy;
    int randomnumber;
    int permute[8] = {0, 1, 2, 3, 4, 5, 6, 7};
    int permutetmp;

    preprocessmaze();
    for (y = 0; y < MAZEHEIGHT; ++y){
      for (x = 0; x < MAZEWIDTH; ++x){
	maze[y][x].obstacle = 1;
	maze[y][x].dfsx = 0;
	maze[y][x].dfsy = 0;
      }
    }
    x = 0;
    y = 0;
    maze[y][x].dfsx = -1;
    maze[y][x].dfsy = -1;
    while (1){
      if (maze[y][x].obstacle)
	maze[y][x].obstacle = 0;
      for (d = 0; d < DIRECTIONS-1; ++d){
	randomnumber = random() % (DIRECTIONS-d);
	permutetmp = permute[randomnumber];
	permute[randomnumber] = permute[DIRECTIONS-1-d];
	permute[DIRECTIONS-1-d] = permutetmp;
      }
      for (dtmp = 0; dtmp < DIRECTIONS; ++dtmp){
	d = permute[dtmp];
	newx = (x + 2*dx[d] + MAZEWIDTH) % MAZEWIDTH;
	newy = (y + 2*dy[d] + MAZEHEIGHT) % MAZEHEIGHT;
	if (maze[y][x].succ[d] != NULL && maze[newy][newx].obstacle){
	  if (maze[(y + dy[d] + MAZEWIDTH) % MAZEWIDTH][(x + dx[d] + MAZEHEIGHT) % MAZEHEIGHT].obstacle)
	    maze[(y + dy[d] + MAZEWIDTH) % MAZEWIDTH][(x + dx[d] + MAZEHEIGHT) % MAZEHEIGHT].obstacle = 0;
	  maze[newy][newx].dfsx = x;
	  maze[newy][newx].dfsy = y;
	  x = newx;
	  y = newy;
	  break;
	}
      }
      if (dtmp == DIRECTIONS){
	if (maze[y][x].dfsx == -1)
	  break;
	newx = maze[y][x].dfsx;
	newy = maze[y][x].dfsy;
	x = newx;
	y = newy;
      }
    }
    while (wallstoremove){
      newx = random() % MAZEWIDTH;
      newy = random() % MAZEHEIGHT;
      if (maze[newy][newx].obstacle){
	maze[newy][newx].obstacle = 0;
	--wallstoremove;
      }
    }
}

void newrandommaze() 
{
  int x, y;
  int goaly, goalx;    
  int starty, startx; 

#ifdef RANDOMSTARTGOAL
  goaly = (random() % ((MAZEHEIGHT + 1) / 2 )) * 2;
  goalx = (random() % ((MAZEWIDTH  + 1) / 2 )) * 2;
  while (1){
    starty = (random() % ((MAZEHEIGHT + 1) / 2 )) * 2;
    startx = (random() % ((MAZEWIDTH  + 1) / 2 )) * 2;
    if (startx != goalx || starty != goaly)
      break;
  }
#endif
  if (maze == NULL){
    maze = (cell **)calloc(MAZEHEIGHT , sizeof(cell *));
    for (y = 0; y < MAZEHEIGHT; ++y)
      maze[y] = (cell *)calloc(MAZEWIDTH, sizeof(cell));
  } 
  for (y = 0; y < MAZEHEIGHT; ++y){
    for (x = 0; x < MAZEWIDTH; ++x){
      maze[y][x].obstacle = (random() % 10000 < 10000 * MAZEDENSITY);
    }
  }
  maze[starty][startx].obstacle = 0;
  maze[goaly ][goalx ].obstacle = 0;
}

int main()
{
  int row,col;
  srandom(time(NULL));
  newdfsmaze(WALLSTOREMOVE);  
  // alternatively: newrandommaze();
  for (col=0; col<MAZEWIDTH+2; col++)
    printf("X");
  printf("\n");
  for (row=0; row<MAZEHEIGHT; row++){
    printf("X");
    for (col=0; col<MAZEWIDTH; col++){
      if (maze[row][col].obstacle){
	printf("X");
      }
      else if (row == starty && col == startx){
	printf("A");
      }	
      else if (row == goaly && col == goalx){
	printf("T");
      }
      else {
	printf(" ");
      }
    }
    printf("X\n");
  }
  for (col=0; col<MAZEWIDTH+2; col++)
    printf("X");
  printf("\n");
  return 0;
}

